/* This is a simple test server program to process a CSTA MakeCall      */
/* request from the CSTA client application (client.cpp).               */
/* Communications are done using HTTP / SOAP / XML.  This is an         */
/* extremely simple case is that the server exits after processing      */
/* a single request.  A real world server would continuously            */
/* process requests indefinitely from multiple clients..                */

#include <stdio.h>
#include <stdlib.h>
#include "rtxsrc/rtxSOAP.h"
#include "rtxsrc/rtxSocket.h"
#include "rtxsrc/rtxStreamMemory.h"
#include "rtxsrc/rtxStreamSocket.h"
#include "rtxsrc/OSRTMemoryInputStream.h"
#include "rtxmlsrc/OSXMLDecodeBuffer.h"
#include "rtxmlsrc/OSXMLEncodeBuffer.h"
#include "csta.h"

#define DEFAULT_URL "http://127.0.0.1:9125"

int main (int argc, char** argv)
{
   OSCTXT       ctxt;
   OSSOAPCONN   soapConn;
   OSOCTET*     ptmpbuf = 0;
   OSIPADDR     ipAddr = OSIPADDR_LOCAL;
   OSRTSOCKET   listenSocket;
   const char*  url = DEFAULT_URL;
   int		i, stat;
   OSBOOL       verbose = FALSE, trace = TRUE;

   /* Process command line arguments */

   if (argc > 1) {
      for (i = 1; i < argc; i++) {
         if (!strcmp (argv[i], "-v")) verbose = TRUE;
         else if (!strcmp (argv[i], "-url")) {
            url = argv[++i];
         }
         else if (!strcmp (argv[i], "-notrace")) trace = FALSE;
         else {
            printf ("usage: server [-v] [-url <url>] [-notrace]\n");
            printf ("   -v    verbose mode: print trace info\n");
            printf ("   -url  bind URL http://host:port\n");
            printf ("   -notrace  do not display trace info\n");
            return 0;
         }
      }
   }

   /* Init context */

   stat = rtXmlInitContext (&ctxt);
   if (0 != stat) {
      printf ("Context initialization failed.\n");
      rtxErrPrint (&ctxt);
      return stat;
   }

   if (verbose) 
      rtxSetDiag (&ctxt, 1);

   rtxSoapInitConn (&soapConn, &ctxt, OSSOAPV1, url);

   /* One-time socket package initialization (required for Windows only) */

   stat = rtxSocketsInit();
   if (0 != stat) {
      printf ("Socket initialization failed.\n");
      rtxErrSetData (&ctxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&ctxt);
      return stat;
   }

   /* Create listener socket */

   if ((stat = rtxSocketCreate (&listenSocket)) != 0) {
      rtxErrSetData (&ctxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&ctxt);
      return stat;
   }

   /* Bind the socket */

   if ((stat = rtxSocketStrToAddr (soapConn.host, &ipAddr)) != 0) {
      rtxErrSetData (&ctxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&ctxt);
      return stat;
   }
   if ((stat = rtxSocketBind (listenSocket, ipAddr, soapConn.port)) != 0) {
      rtxErrSetData (&ctxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&ctxt);
      return stat;
   }

   /* Wait for connection */

   if ((stat = rtxSocketListen (listenSocket, 1)) != 0) {
      rtxErrSetData (&ctxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&ctxt);
      return stat;
   }

   if (trace) printf ("Wait for connection...\n");

   stat = rtxSoapAcceptConn (listenSocket, &soapConn);
   if (0 != stat) {
      rtxErrSetData (&ctxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&ctxt);
      return stat;
   }

   /* Wait for request */

   stat = rtxSoapRecvHttp (&soapConn);
   if (stat != 0) {
      rtxErrPrint (&ctxt);
      return stat;
   }
   if (trace) printf ("HTTP request received\n");

   /* Read HTTP request into memory.  This is the XML message to be     */
   /* decoded..                                                         */

   stat = rtxSoapRecvHttpContent (&soapConn, &ptmpbuf);
   if (stat != 0) {
      rtxErrPrint (&ctxt);
      return stat;
   }

   // Decode make call request message

   OSRTMemoryInputStream in (ptmpbuf, soapConn.contentLength);
   OSXMLDecodeBuffer decodeBuffer (in);

   if (decodeBuffer.getStatus ()) {
      decodeBuffer.printErrorInfo();
      return decodeBuffer.getStatus ();
   }
   OSCTXT* pDecCtxt = decodeBuffer.getCtxtPtr();
   OSXMLElemDescr soapElem;

   // First 2 elements should be SOAP envelope and body

   for (int j = 0; j < 2; j++) {
      stat = rtXmlpGetNextElem (pDecCtxt, &soapElem, -1);
      if (0 == stat) {
         printf ("soapElem = '%.*s'\n", 
                 soapElem.localName.length, soapElem.localName.value);
      }
      else {
         printf ("rtXmlpGetNextElem failed\n");
         decodeBuffer.printErrorInfo();
         return stat;
      }
   }

   // Decode SOAP body content

   MakeCall_ELEM value;
   MakeCall_CC pdu (decodeBuffer, value);

   decodeBuffer.setDiag (verbose);

   stat = pdu.decode();

   if (0 == stat) {
      if (trace) {
         printf ("decoded XML message:\n");
         pdu.print ("MakeCall");
         printf ("\n");
      }
   }
   else {
      printf ("Decoding failed\n");
      decodeBuffer.printErrorInfo();
      return stat;
   }

   /* Prepare a response to the request */

   MakeCallResponse_ELEM data;
   OSXMLEncodeBuffer buffer;
   MakeCallResponse_CC respPDU (buffer, data);

   buffer.setDiag (verbose);

   // Populate structure of generated type

   MakeCallResponse_ELEM* pvalue = &data;
   pvalue->callingDevice.t = 1;
   pvalue->callingDevice.u.deviceID = new LocalDeviceID();
   pvalue->callingDevice.u.deviceID->value.set_number (new Number("1234"));
   pvalue->callingDevice.u.deviceID->typeOfNumber
      .setValue (LocalDeviceID_typeOfNumber::dialingNumber);
   pvalue->callingDevice.u.deviceID->m.typeOfNumberPresent = 1;
   pvalue->callingDevice.u.deviceID->mediaClass.voiceBit = 1;
   pvalue->callingDevice.u.deviceID->mediaClass.dataBit = 1;
   pvalue->callingDevice.u.deviceID->mediaClass.imageBit = 0;
   pvalue->callingDevice.u.deviceID->mediaClass.audioBit = 0;
   pvalue->callingDevice.u.deviceID->mediaClass.otherBit = 0;
   pvalue->callingDevice.u.deviceID->mediaClass.notKnownBit = 0;
   pvalue->callingDevice.u.deviceID->mediaClass.chatBit = 1;
   pvalue->callingDevice.u.deviceID->mediaClass.emailBit = 0;
   pvalue->callingDevice.u.deviceID->mediaClass.messageBit = 0;
   pvalue->callingDevice.u.deviceID->mediaClass.imBit = 0;
   pvalue->callingDevice.u.deviceID->mediaClass.smsBit = 1;
   pvalue->callingDevice.u.deviceID->mediaClass.mmsBit = 1;
   pvalue->callingDevice.u.deviceID->mediaClass._extItems = 0;

   pvalue->callingDevice.u.deviceID->m.mediaClassPresent = 1;
   pvalue->callingDevice.u.deviceID->connectionRate
       = OSI64CONST(6121986096669042600);
   pvalue->callingDevice.u.deviceID->m.connectionRatePresent = 1;
   pvalue->callingDevice.u.deviceID->bitRate
      .setValue (LocalDeviceID_bitRate::constant);
   pvalue->callingDevice.u.deviceID->m.bitRatePresent = 1;
   pvalue->callingDevice.u.deviceID->delayTolerance
       = OSI64CONST(-5879910229128678607);
   pvalue->callingDevice.u.deviceID->m.delayTolerancePresent = 1;
   pvalue->callingDevice.u.deviceID->switchingSubDomainCCIEType
      .setValue (LocalDeviceID_switchingSubDomainCCIEType::isdn);
   pvalue->callingDevice.u.deviceID->m.switchingSubDomainCCIETypePresent = 1;
   { static const OSOCTET testdata[] = { 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f };
   pvalue->callingDevice.u.deviceID->switchingSubDomainInformationElements
      .setValue (sizeof(testdata), testdata);
   }
   pvalue->callingDevice.u.deviceID
      ->m.switchingSubDomainInformationElementsPresent = 1;

   // Add XML document and SOAP headers to encode buffer

   buffer.addXMLHeader ();

   stat = rtXmlEncStartSoapElems (buffer.getCtxtPtr(), OSSOAPBODY);

   // Encode SOAP body content

   stat = respPDU.encode();

   // Close SOAP body and envelope elements

   if (0 == stat) {
      stat = rtXmlEncEndSoapElems (buffer.getCtxtPtr(), OSSOAPBODY);
   }

   rtXmlEncEndDocument (buffer.getCtxtPtr());

   if (0 == stat) {
      if (trace) {
         printf ("encoded XML message:\n");
         printf ((const char*)buffer.getMsgPtr());
         printf ("\n");
      }
   }
   else {
      printf ("Encoding failed\n");
      buffer.printErrorInfo();
      return stat;
   }

   /* Send response */

   stat = rtxSoapSendHttp (&soapConn, buffer.getMsgPtr());
   if (stat != 0) {
      buffer.printErrorInfo();
      return stat;
   }
   if (trace) printf ("HTTP response sent\n");

   /* Close connection */

   rtxStreamClose (&ctxt);
   rtxFreeContext (&ctxt);

   return stat;
}
